package io.data2viz.charts.chart

import io.data2viz.charts.chart.mark.TooltipPosition
import io.data2viz.charts.core.*
import io.data2viz.charts.layout.DrawingZone
import io.data2viz.viz.RichTextBuilder

/**
 * The [Chart] tooltip visual component.
 */
public interface Tooltip<DOMAIN> {

    /**
     * Set the title of the tooltip, defaults to the series name (if any), but you can choose to use another title.
     * For more complex tooltip layout use the [tooltipBuilder] lambda.
     */
    public var title: (Datum<DOMAIN>.() -> String)?

    /**
     * Set the [Datum] formatter, by default, check if the Datum.domain object implements the [Labelled] interface and
     * displays Datum.domain.label, else the formatter will try to display infos based on the dimension used.
     *
     * @see [Labelled]
     */
    public var formatter: (Datum<DOMAIN>.() -> String)?


    /**
     * Define a "tooltip builder", a lambda that takes several parameters:
     *
     * @param selectedData: the [SelectedDatum] object on which the tooltip "applies"
     * @param ttp: a (nullable) [TooltipPosition] storing where to place the tooltip on the [DrawingZone]
     * @param defaultTitle: a nullable [String], the default title of your tooltip (if any),
     *  you can set your own title by providing a [title] lambda
     * @param defaultText: a nullable [String], the default text content of your tooltip (if any),
     *  you can set your own content by providing a [formatter] lambda
     * @param displayRatio: the tooltip zoom-out ratio, defaults to 1.0, you may get a smaller ratio (0.7) if the
     *  tooltip is too big for the chart, in this case, use this ratio to reduce your layout and/or your font size.
     * @param defaultFont: the tooltip's [Font] that comes from the actual theme used
     * @param drawingZone: the [DrawingZone] on which to draw the custom tooltip
     * @return a [RichTextBuilder] (or null), then the tooltip itself will size and display accordingly.
     */
    public var tooltipBuilder: (
        selectedData: SelectedDatum<DOMAIN>?,
        ttp: TooltipPosition?,
        defaultTitle: String?,
        defaultText: String?,
        displayRatio: Double,
        defaultFont: Font,
        drawingZone: DrawingZone
    ) -> RichTextBuilder?

}

