package io.data2viz.charts.chart.mark

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.mark.internal.BarMarkImpl
import io.data2viz.charts.config.AxisConfig
import io.data2viz.charts.config.EventConfig
import io.data2viz.charts.config.MarkConfig
import io.data2viz.charts.core.Dataset
import io.data2viz.charts.core.Datum
import io.data2viz.charts.dimension.Dimension
import io.data2viz.charts.dimension.Discrete
import io.data2viz.charts.dimension.Quantitative

/**
 * Instantiate a new [BarMark].
 *
 * See [Bar chart](https://en.wikipedia.org/wiki/Bar_chart) for more info.
 *
 * @param dataset Your chart's [Dataset]
 * @param x The [Quantitative] dimension for the X axis
 * @param y The [Discrete] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 * @param eventConfig The [EventConfig] to configure this [Mark]'s events
 * @param config The [MarkConfig] to configure this [Mark]
 * @param xAxisConfig The [AxisConfig] to configure this mark's X axis
 * @param yAxisConfig The [AxisConfig] to configure this mark's Y axis
 */
public fun <DOMAIN, YVAL> barMark(
    markIndex: Int,
    dataset: Dataset<DOMAIN>,
    x: Quantitative<DOMAIN>,
    y: Discrete<DOMAIN, YVAL>,
    init: BarMark<DOMAIN, Double?, YVAL>.() -> Unit,
    eventConfig: EventConfig,
    config: MarkConfig,
    xAxisConfig: AxisConfig,
    yAxisConfig: AxisConfig
): BarMark<DOMAIN, Double?, YVAL> = BarMarkImpl(markIndex, dataset, x, y, false, init, eventConfig, config, xAxisConfig, yAxisConfig)


/**
 * Instantiate a new [BarMark].
 *
 * See [Bar chart](https://en.wikipedia.org/wiki/Bar_chart) for more info.
 *
 * @param dataset Your chart's [Dataset]
 * @param x The [Discrete] dimension for the X axis
 * @param y The [Quantitative] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 * @param eventConfig The [EventConfig] to configure this [Mark]'s events
 * @param config The [MarkConfig] to configure this [Mark]
 * @param xAxisConfig The [AxisConfig] to configure this mark's X axis
 * @param yAxisConfig The [AxisConfig] to configure this mark's Y axis
 */
public fun <DOMAIN, XVAL> barMark(
    markIndex: Int,
    dataset: Dataset<DOMAIN>,
    x: Discrete<DOMAIN, XVAL>,
    y: Quantitative<DOMAIN>,
    init: BarMark<DOMAIN, XVAL, Double?>.() -> Unit,
    eventConfig: EventConfig,
    config: MarkConfig,
    xAxisConfig: AxisConfig,
    yAxisConfig: AxisConfig
): BarMark<DOMAIN, XVAL, Double?> = BarMarkImpl(markIndex, dataset, x, y, true, init, eventConfig, config, xAxisConfig, yAxisConfig)

/**
 * Add a horizontal [BarMark] to the current [Chart].
 *
 * See [Bar chart](https://en.wikipedia.org/wiki/Bar_chart) for more info.
 *
 * @param x The [Quantitative] dimension for the X axis
 * @param y The [Discrete] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 */
public fun <DOMAIN, YVAL> Chart<DOMAIN>.bar(
    x: Quantitative<DOMAIN>,
    y: Discrete<DOMAIN, YVAL>,
    init: BarMark<DOMAIN, Double?, YVAL>.() -> Unit = {}
): Unit = addMark { markIndex -> barMark(markIndex, dataset, x, y, init, config.events, config.mark, config.x, config.y) }

/**
 * Add a vertical [BarMark] to the current [Chart].
 *
 * See [Bar chart](https://en.wikipedia.org/wiki/Bar_chart) for more info.
 *
 * @param x The [Discrete] dimension for the X axis
 * @param y The [Quantitative] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 */
public fun <DOMAIN, XVAL> Chart<DOMAIN>.bar(
    x: Discrete<DOMAIN, XVAL>,
    y: Quantitative<DOMAIN>,
    init: BarMark<DOMAIN, XVAL, Double?>.() -> Unit = {}
): Unit = addMark { markIndex -> barMark(markIndex, dataset, x, y, init, config.events, config.mark, config.x, config.y) }


/**
 * Draw horizontal bars for a given [Dataset] and X and Y [Dimension]s.
 */
public interface BarMark<DOMAIN, XVAL, YVAL> :
    HasFill<DOMAIN>,
    HasLabel<DOMAIN>,
    HasStrokeColor<DOMAIN>,
    HasStrokeWidth<DOMAIN>,
    HasStack,
    HasCustomTooltip<DOMAIN>,
    HasMarkDecorator<DOMAIN>,
    HasHighlightDecorator<DOMAIN>,
    HasSelectionDecorator<DOMAIN>,
    HasEventZones<DOMAIN>,
    HasLegendDecorator<DOMAIN>,
    HasHighlightMode,
    HasSelectionMode,
    MarkXY<DOMAIN, XVAL, YVAL> {

    /**
     * The "base line" [Double] value of each bar, default to the constant value 0.
     * Use a variable [baseline] value to have "floating" bars.
     */
    public var baseline: Datum<DOMAIN>.()-> Double

}


@Deprecated(
    "Replace with BarMark.",
    ReplaceWith("BarMark<DOMAIN, YVAL>", "io.data2viz.charts.chart.mark.BarMark")
)
public typealias Bar<DOMAIN, YVAL> = BarMark<DOMAIN, Double?, YVAL>
