package io.data2viz.charts.chart.mark

import io.data2viz.charts.core.Datum
import io.data2viz.charts.core.ExperimentalCharts
import io.data2viz.charts.event.EventZone
import io.data2viz.charts.layout.DrawingZone
import io.data2viz.geom.Point

/**
 * A decorator lambda is called on each point, use this to draw custom decorations for your signs.
 * It can be used on mark, selection, highlight.
 *
 * @param datum the current element
 * @param position the expected position where the sign should appear (relative to the [drawingZone] coordinates)
 * @param drawingZone the drawing zone on which to draw the decorator
 */
public typealias DatumDecorator<DOMAIN> = (datum: Datum<DOMAIN>, position: Point, drawingZone: DrawingZone) -> Unit

/**
 * Decorator lambda specific to the [BoxPlot], warning this is experimental and can be deprecated to a more robust
 * mark drawer.
 */
@ExperimentalCharts
public typealias BoxPlotDecorator<DOMAIN> = (datum: Datum<DOMAIN>, context: BoxPlotContext, drawingZone: DrawingZone) -> Unit


public interface HasBoxPlotDecorator<DOMAIN> {
    public val defaultMarkDrawer: BoxPlotDecorator<DOMAIN>
    public var markDrawer: BoxPlotDecorator<DOMAIN>
}
public interface HasBoxPlotHighlightDecorator<DOMAIN> {
    public val defaultHighlightDrawer: BoxPlotDecorator<DOMAIN>
    public var highlightDrawer: BoxPlotDecorator<DOMAIN>
}
public interface HasBoxPlotSelectionDecorator<DOMAIN> {
    public val defaultSelectionDrawer: BoxPlotDecorator<DOMAIN>
    public var selectionDrawer: BoxPlotDecorator<DOMAIN>
}

/**
 * A EventZoneBuilder lambda is called on each point, use this to draw customs [EventZone] for your signs.
 *
 * @param datum the current element
 * @param position the expected position where the sign should appear (relative to the [drawingZone] coordinates)
 * @param drawingZone the drawing zone (screen space) on which the EventZone should be build
 */
public typealias EventZoneBuilder<DOMAIN> = (datum: Datum<DOMAIN>, position: Point, drawingZone: DrawingZone) -> List<EventZone<DOMAIN>>

/**
 * A mark that implements HasMarkDecorator allows to define a [DatumDecorator] for each of its [Datum]s.
 */
public interface HasMarkDecorator<DOMAIN> {

    /**
     * The default mark decorator for this [Mark], if you use a custom [DatumDecorator] you may want to call
     * defaultMarkDecorator(datum, x, y, drawingZone) to draw the expected sign.
     */
    public val defaultMarkDecorator: DatumDecorator<DOMAIN>

    /**
     * The mark decorator lambda will be called on each point, use this to draw custom decorations for your signs.
     * Initialized to defaultMarkDecorator.
     */
    public var markDecorator: DatumDecorator<DOMAIN>
}

/**
 * A mark that implements HasMarkDecorator allows to define a [DatumDecorator] to highlight each of its [Datum]s.
 */
public interface HasHighlightDecorator<DOMAIN> {

    /**
     * The default highlight decorator for this [Mark], if you use a custom [DatumDecorator] you may want to call
     * defaultHighlightDecorator(datum, x, y, drawingZone) to draw the expected sign.
     */
    public val defaultHighlightDecorator: DatumDecorator<DOMAIN>

    /**
     * The highlight decorator lambda will be called on each sign, use this to draw custom decorations
     * on a highlighted sign.
     * Initialized to defaultHighlightDecorator.
     */
    public var highlightDecorator: DatumDecorator<DOMAIN>
}

/**
 * A mark that implements HasMarkDecorator allows to define a [DatumDecorator] to select each of its [Datum]s.
 */
public interface HasSelectionDecorator<DOMAIN> {

    /**
     * The default selection decorator for this [Mark], if you use a custom [DatumDecorator] you may want to call
     * defaultSelectionDecorator(datum, x, y, drawingZone) to draw the expected sign.
     */
    public val defaultSelectionDecorator: DatumDecorator<DOMAIN>

    /**
     * The selection decorator lambda will be called on each sign, use this to draw custom decorations
     * on a selected sign.
     * Initialized to defaultSelectionDecorator.
     */
    public var selectionDecorator: DatumDecorator<DOMAIN>
}

/**
 * A mark that implements HasLegendDecorator allows to define a custom legend decorator.
 */
public interface HasLegendDecorator<DOMAIN> {

    /**
     * The default legend decorator for a given [Datum].
     */
    public val defaultLegendDecorator: DatumDecorator<DOMAIN>

    /**
     * The legend decorator lambda will be called on each series, use this to draw custom decorations for your series.
     * Initialized to defaultLegendDecorator.
     */
    public var legendDecorator: DatumDecorator<DOMAIN>
}

/**
 * A mark that implements HasEventZoneBuilder allows to define custom [EventZone].
 */
public interface HasEventZones<DOMAIN> {

    /**
     * The default [EventZone] builder for a given [Datum].
     */
    public val defaultEventZoneBuilder: EventZoneBuilder<DOMAIN>

    /**
     * The eventZoneBuilder lambda will be called on each sign, use this if you want to return custom [EventZone].
     * Initialized to defaultEventZoneBuilder.
     */
    public var eventZoneBuilder: EventZoneBuilder<DOMAIN>
}

