package io.data2viz.charts.config

import io.data2viz.charts.axis.AxisLabelLayout
import io.data2viz.charts.axis.TickSkip
import io.data2viz.charts.core.*
import io.data2viz.charts.core.Font
import io.data2viz.charts.core.Stroke
import io.data2viz.color.Color
import io.data2viz.math.Angle
import io.data2viz.math.Percent
import io.data2viz.scale.FirstLastRange

/**
 * The configuration interface for an axis.
 */
public interface AxisConfig: Font, Stroke, TickConfig {

    /**
     * The [Orientation] of this axis configuration.
     * This allows to store different configuration for different types of axes (horizontal, vertical...).
     */
    public val orientation: Orientation

    /**
     * The rotation of every labels as an [Angle] (generally 0.deg <= labelRotation <= 90.deg)
     * LabelRotation is null (by default), and the [AxisRenderer] use "smart rotation" to layout labels.
     * Set an [Angle] to rotate labels.
     *
     * Note: labels can only rotate in one direction.
     */
    public var labelRotation: Angle?

    /**
     * Returns a copy of the current axis configuration
     */
    public fun copy(font: Font, stroke: Stroke): AxisConfig

    /**
     * Show or hide the whole axis component.
     */
    public var enableAxis: Boolean

    /**
     * Show or hide the axis line.
     */
    public var enableAxisLine: Boolean

    /**
     * When autofit is enabled, the start and end of the axis is rounded to "nice values".
     * When autofit is disabled, the start and end of the axis are the min/max of the dataset values.
     *
     * Note : autofit is disabled on discrete axes, or if the user input a start/end value.
     */
    public var autoFit: Boolean

    /**
     * Show or hide the axis ticks.
     */
    public var enableTicks: Boolean

    /**
     * Show or hide the axis ticks labels.
     */
    public var enableTicksLabels: Boolean

    /**
     * Layout of the ticks labels over the axis line (below, over...).
     */
    @Deprecated(message = "Remove axisLabelLayout, the axis now manage its own smart layout.")
    public var axisLabelLayout: AxisLabelLayout

    /**
     * Tick-sipping behavior for a discrete axis.
     * @see tickCount
     */
    @Deprecated(message = "Remove discreteTickSkip, the axis now use a smart skip behavior.")
    public var discreteTickSkip: TickSkip

    /**
     * The expected tick count for the [Axis], null for activating "smart tick count".
     * Note: this is only an indicative count, sometimes the [Axis] will use more ticks to produce a nice result.
     */
    public var tickCount: Int?

    /**
     * Show / hide axis title.
     */
    public var enableTitle: Boolean

    /**
     * The padding between the tick and its label.
     */
    public var tickPadding: Double

    /**
     * The size (length) of a tick.
     */
    public var tickSize: Double

    /**
     * The ratio of the drawable zone on a discrete axis.
     *
     * For example, a barchart with a bandWidth of 80% will have 10% margin on each side of each bar.
     */
    public var bandwidthRatio: Percent

    /**
     * Enable bandwith.
     * @see [Bandwidth]
     */
//    public var enableBandwidth: Bandwidth

    /**
     * Show hide gridlines (lines drawn on ticks under the [Mark]s).
     */
    public var enableGridLines: Boolean

    /**
     * The color used to draw the grid lines.
     */
    public var gridLinesColor: Color

    /**
     * The thickness of the grid lines.
     */
    public var gridLinesWidth: Double

    /**
     * The "dash style" of the grid lines, use null for a solid line.
     */
    public var gridLinesDashed: DoubleArray?

    /**
     * Show hide alternate background (colored bands, drawn on ticks under the [Mark]s).
     */
    public var enableAlternateBackground: Boolean

    /**
     * The color of the alternate background bands.
     */
    public var alternateBackgroundColor: Color

    /**
     * Expected position for the axis relative to the main charting zone.
     */
    public var layoutPosition: LayoutPosition


}
