package io.data2viz.charts.config

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.Tooltip
import io.data2viz.charts.chart.internal.CursorImpl
import io.data2viz.charts.chart.mark.Axis
import io.data2viz.charts.chart.mark.Mark
import io.data2viz.charts.config.configs.debugConfig
import io.data2viz.charts.config.configs.greenConfig
import io.data2viz.charts.config.configs.lightConfig
import io.data2viz.charts.config.internal.*
import io.data2viz.charts.core.ExperimentalCharts
import io.data2viz.charts.core.Font
import io.data2viz.charts.core.Stroke
import io.data2viz.charts.core.Zoom


/**
 * Build a chart configuration.
 */
public fun configuration(
    init: ChartConfig.() -> Unit = {}
): ChartConfig =
    ChartConfigImpl().apply(init)

// TODO manage text alignment
// TODO linear (continuous) fills and strokes for according scales

/**
 * The ChartConfig is the central point of the configuration that holds a lot of configuration properties.
 *
 * Some of them are defined as root configuration properties. Other are grouped into specific scoped configurations
 * (cursor, events, mark, selection, tooltip, zoom, etc.)
 *
 *  *
 */
public interface ChartConfig:
    Font,
    Stroke,
    ChartBasicConfig,
    Config<ChartConfig>
{

    /**
     * The cursor configuration
     */
    public val cursor: CursorConfig

    /**
     * The events configuration
     */
    public val events: EventConfig

    /**
     * The legend configuration
     */
    public val legend: LegendConfig

    /**
     * The global mark configuration
     */
    public val mark: MarkConfig

    /**
     * The selection configuration
     */
    public val selection: SelectionConfig

    /**
     * The selection configuration when adding elements
     */
    public val selectionAdd: SelectionConfig

    /**
     * The selection configuration when removing elements
     */
    public val selectionRemove: SelectionConfig

    /**
     * The tooltip configuration
     */
    public val tooltip: TooltipConfig

    /**
     * The x axis configuration
     */
    public val x: AxisConfig

    /**
     * The y axis configuration
     */
    public val y: AxisConfig

    /**
     * The zoom configuration
     */
    public val zoom: ZoomConfig

    /**
     * The title configuration
     */
    public val title: TitleConfig

    /**
     * The [MarkConfig] initialization lambda, use it to customize your [Mark]s.
     */
    public fun mark(init: MarkConfig.() -> Unit)

    /**
     * The [EventConfig] initialization lambda, use it to customize your events behaviors.
     */
    public fun events(init: EventConfig.() -> Unit)

    /**
     * The [LegendConfig] initialization lambda, use it to customize your [Legend].
     */
    @ExperimentalCharts
    public fun legend(init: LegendConfig.() -> Unit)

    /**
     * The [ChartBasicConfig] initialization lambda, use it to customize your [Chart].
     */
    public fun chart(init: ChartBasicConfig.() -> Unit)

    /**
     * The [TooltipConfig] initialization lambda, , use it to customize your [Tooltip].
     */
    public fun tooltip(init: TooltipConfig.() -> Unit)

    /**
     * The [TitleConfig] initialization lambda, , use it to customize your [Title].
     */
    public fun title(init: TitleConfig.() -> Unit)

    /**
     * The [CursorConfigImpl] initialization lambda, use it to customize your [CursorImpl].
     */
    public fun cursor(init: CursorConfig.() -> Unit)

    /**
     * The [ZoomConfig] initialization lambda, use it to customize your [Zoom].
     */
    public fun zoom(init: ZoomConfig.() -> Unit)

    /**
     * The [SelectionConfig] initialization lambda, use it to customize style your selection layer.
     */
    public fun selection(init: SelectionConfig.() -> Unit)

    /**
     * The X-[AxisConfig] initialization lambda, use it to customize your X [Axis].
     */
    public fun xAxis(init: AxisConfig.() -> Unit)

    /**
     * The Y-[AxisConfig] initialization lambda, use it to customize your Y [Axis].
     */
    public fun yAxis(init: AxisConfig.() -> Unit)


    /**
     * This is a library of default configurations, you can create your own configuration or use one of these.
     */
    public companion object {

        /**
         * The default [ChartConfig], may be used in any situation.
         */
        public val default: ChartConfig = configuration()


        /**
         * Light themed [ChartConfig], use light colors and less space than other configurations.
         */
        public val light: ChartConfig = lightConfig

        /**
         * Debug [ChartConfig], strong lines, vibrant colors, should not be used in production but useful for debugging
         * when creating custom components.
         */
        public val debug: ChartConfig = debugConfig

        /**
         * Green themed [ChartConfig].
         */
        public val green: ChartConfig = greenConfig
    }

}
