package io.data2viz.charts.config

import io.data2viz.charts.chart.mark.LineMark
import io.data2viz.charts.chart.mark.Mark
import io.data2viz.charts.chart.mark.MarkCurves
import io.data2viz.charts.chart.mark.PathDrawer
import io.data2viz.charts.core.Font
import io.data2viz.charts.core.GridPosition
import io.data2viz.charts.core.MarkLabel
import io.data2viz.charts.core.Stroke
import io.data2viz.color.*
import io.data2viz.math.pct
import io.data2viz.shape.Symbols

// TODO
internal fun ColorOrGradient?.brighten(strength: Double = 1.0): Color? {
    return when (this) {
        is Color    -> this.brighten(strength)
        is Gradient -> null
        else        -> null
    }
}

/**
 * [Mark] configuration.
 */
public data class MarkConfig internal constructor(
    private val font: Font,
    private val defaultStroke: Stroke,

    /**
     * The type of curves the [Mark] will draw in the form of a [PathDrawer].
     *
     * @see MarkCurves
     */
    public var curveType: PathDrawer = MarkCurves.Straight,

    /**
     * The line dash style, set to null for a solid line.
     */
    public var strokeLine: DoubleArray? = null,

    /**
     * Show / hide markers, for example a [LineMark] with showMarkers will draw a sign at each point of the line.
     */
    public var showMarkers: Boolean = false,

    /**
     * Default marker size.
     */
    public var markersSize: Double = 15.0,

    /**
     * List of different [Symbols] for drawing markers, when having multiple series, each series will use a different one.
     */
    public var markers: List<Symbols> = listOf(
        Symbols.Circle,
        Symbols.Square,
        Symbols.Diamond,
        Symbols.Triangle,
        Symbols.Cross,
        Symbols.Star,
        Symbols.Wye
    ),

    /**
     * List of different stroke [Color], when having multiple series, each series will use a different one.
     */
    public var strokeColors: List<Color> = listOf(
        "#586ADA".col,
        "#CD3D64".col,
        "#BB5506".col,
        "#09815C".col,
        "#067AB8".col,
        "#C44C35".col,
        "#A44FB5".col,
        "#825FC3".col,
        "#687386".col
    ),

    /**
     * List of different filling [Color], when having multiple series, each series will use a different one.
     */
    public var fills: List<Color> = strokeColors.map{ it.desaturate(1.0).brighten(1.5).withAlpha(70.pct) },

    /**
     * List of different stroke [Color] for Highlighted elements.
     */
    public var strokeColorsHighlight: List<Color> =  strokeColors.toList(),

    /**
     * List of different fill [Color] for Highlighted elements.
     */
    public var fillsHighlight: List<Color> = fills.map{ it.desaturate(0.5).brighten(1.0) },

    /**
     * StrokeWidth for Highlighted elements
     */
    public var strokeWidthHighlight: Double? = (defaultStroke.strokeWidth ?: .0) + 2.0,

    /**
     * List of different stroke [Color] for Selected elements.
     */
    public var strokeColorsSelect: List<Color>  = listOf(Colors.Web.orangered),

    /**
     * List of different fill [Color] for Selected elements..
     */
    public var fillsSelect: List<Color> = listOf(Colors.Web.orangered.desaturate(1.0).brighten(1.5)),

    /**
     * StrokeWidth for Selected elements
     */
    public var strokeWidthSelect: Double? = (defaultStroke.strokeWidth ?: .0) + 1.0,

    /**
     * Display label over marks: display X and or Y values or NONE.
     */
    public var label: MarkLabel = MarkLabel.None,

    /**
     * The position of the label relative to the [Mark] screen position.
     */
    public var labelPosition: GridPosition = GridPosition.Top,

    /**
     * The distance of the label from the [Mark] screen position.
     */
    public var labelDistance: Double = 10.0
):
    Font by font,
    Stroke by defaultStroke
{

    @Deprecated("Use 'strokeColors'. Will be removed in a future release", ReplaceWith("strokeColors"))
    public var discreteStrokes: List<Color> = strokeColors

    @Deprecated("Use 'fills'. Will be removed in a future release", ReplaceWith("fills"))
    public var discreteFills: List<Color> = fills


    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other == null || this::class != other::class) return false

        other as MarkConfig

        if (defaultStroke != other.defaultStroke) return false
        if (curveType != other.curveType) return false
        if (strokeLine != null) {
            if (other.strokeLine == null) return false
            if (!strokeLine.contentEquals(other.strokeLine)) return false
        } else if (other.strokeLine != null) return false
        if (showMarkers != other.showMarkers) return false
        if (markersSize != other.markersSize) return false
        if (markers != other.markers) return false
        if (strokeColors != other.strokeColors) return false
        if (fills != other.fills) return false
        if (strokeColorsHighlight != other.strokeColorsHighlight) return false
        if (fillsHighlight != other.fillsHighlight) return false
        if (strokeWidthHighlight != other.strokeWidthHighlight) return false
        if (strokeColorsSelect != other.strokeColorsSelect) return false
        if (fillsSelect != other.fillsSelect) return false
        if (strokeWidthSelect != other.strokeWidthSelect) return false

        return true
    }

    override fun hashCode(): Int {
        var result = defaultStroke.hashCode()
        result = 31 * result + curveType.hashCode()
        result = 31 * result + (strokeLine?.contentHashCode() ?: 0)
        result = 31 * result + showMarkers.hashCode()
        result = 31 * result + markersSize.hashCode()
        result = 31 * result + markers.hashCode()
        result = 31 * result + strokeColors.hashCode()
        result = 31 * result + fills.hashCode()
        result = 31 * result + strokeColorsHighlight.hashCode()
        result = 31 * result + fillsHighlight.hashCode()
        result = 31 * result + strokeWidthHighlight.hashCode()
        result = 31 * result + strokeColorsSelect.hashCode()
        result = 31 * result + fillsSelect.hashCode()
        result = 31 * result + strokeWidthHighlight.hashCode()
        return result
    }
}
