package io.data2viz.charts.core

import io.data2viz.charts.chart.Chart
import io.data2viz.geom.Point
import io.data2viz.geom.Rect
import io.data2viz.geom.RectGeom

/**
 * A [Zone] shaped as a rectangle, used to indicate interactive zones in a [Chart].
 *
 * @see Zone
 */
public data class RectZone(
    /**
     * This [RectZone] top-left coordinates relative to the main charting zone.
     */
    val topLeft: Point,

    /**
     * This [RectZone] bottom-right coordinates relative to the main charting zone.
     */
    val bottomRight: Point

) : Zone {

    public constructor(rect: Rect) : this(rect.topLeft, rect.bottomRight)
    public constructor(rect: Rect, offset: Point) : this(rect.topLeft + offset, rect.bottomRight + offset)

    override fun contains(point: Point): Boolean {
        return point.x >= topLeft.x && point.x <= bottomRight.x && point.y >= topLeft.y && point.y <= bottomRight.y
    }


    override fun getAnchorPoint(layoutPosition: LayoutPosition): Point {
        /**
         * Note that when asking for a "bottom" anchor we still give the top one cause we usually don't want to stick
         * tooltips on the bottom of a RectZone (think of a barchart...)
         */
        return when (layoutPosition) {
            LayoutPosition.Top      -> Point((topLeft.x + bottomRight.x) / 2.0, topLeft.y)
            LayoutPosition.Right    -> Point(bottomRight.x, (topLeft.y + bottomRight.y) / 2.0)
            LayoutPosition.Left     -> Point(topLeft.x, (topLeft.y + bottomRight.y) / 2.0)
            LayoutPosition.Bottom   -> Point((topLeft.x + bottomRight.x) / 2.0, bottomRight.y)
        }
    }

    override val bounds: RectGeom = RectGeom(topLeft, bottomRight)
    override val center: Point
        get() = (topLeft + bottomRight) / 2.0
}
