package io.data2viz.charts.core

import io.data2viz.charts.logging.KLogging
import io.data2viz.scale.ContinuousScale

private val logger = KLogging.logger("io.data2viz.charts.chart")

/**
 * A Zoom object stores information of the current zoom on X and Y directions.
 * @see DirectedZoom
 */
public data class Zoom(

    /**
     * The zoom on the X direction.
     */
    val xZoom: DirectedZoom = DirectedZoom(),

    /**
     * The zoom on the Y direction.
     */
    val yZoom: DirectedZoom = DirectedZoom(),

    ) {

    /**
     * Get the same zoom but with a constraint on the X-axis.
     * It means that the modifier for the X-dimension will always be IDENTITY.
     */
    public fun constraintOnX(): Zoom = Zoom(xZoom, DirectedZoom.IDENTITY)

    /**
     * Get the same zoom but with a constraint on the X-axis.
     * It means that the modifier for the X-dimension will always be IDENTITY.
     */
    public fun constraintOnY(): Zoom = Zoom(DirectedZoom.IDENTITY, yZoom)

    /**
     * Give access to [Zoom.IDENTITY], the default zoom.
     */
    public companion object {

        /**
         * Zoom with a ratio of 1.0 and no offset for both X and Y directions.
         */
        public val IDENTITY: Zoom = Zoom()
    }
}


/**
 * A Zoom on a specific direction (it could be x, y, radial).
 * @see Zoom
 */
public data class DirectedZoom (

    // TODO require > .0 ?
    /**
     * The zoom factor on this direction.
     */
    val ratio: Double = 1.0,

    /**
     * The translation (in pixels) on this direction.
     */
    val offset: Double = .0
) {

    /**
     * Give access to [DirectedZoom.IDENTITY], the default directed zoom.
     */
    public companion object {

        /**
         * Zoom with a ratio of 1.0 and no offset.
         */
        public val IDENTITY: DirectedZoom = DirectedZoom()
    }

    /**
     * Compute a new [DirectedZoom] by applying a zoom ratio to the current one.
     */
    internal fun scale(ratio: Double): DirectedZoom =
        copy(
            ratio = this.ratio * ratio,
            offset = this.offset * ratio,
        )


    /**
     * Compute a new [DirectedZoom] by "offsetting" the current one with the given translation value.
     */
    internal fun translate(translation: Double): DirectedZoom = copy(offset = offset + translation )

    internal fun apply(x: Double): Double = x * ratio + offset
    internal fun invert(d: Double): Double = (d - offset) / ratio

    /**
     * Return a copy of the scale with the domain modified to represent the pan and zoom
     */
    internal fun <VALUE> rescale(scale: ContinuousScale<VALUE,Double>): ContinuousScale<VALUE,Double> {
        if (scale.range.size < 2){
            logger.warn("scale range is not set:: $scale")
            return scale
        }
        return (scale.copy() as ContinuousScale<VALUE, Double>).also { sc ->
            sc.domain = sc.range
                .map { invert(it) }
                .map { scale.invert(it) }
        }
    }
}

/**
 * Compute a DirectedZoom for a scale, based on the "origin scale".
 * The origin scale must not be zoomed, it's DirectedZoom applied has a ratio of 1.0 and an offset of 0.
 *
 * @return the DirectedZoom to transform the original scale to the current scale so result.rescale(originScale) == this
 */
internal fun <VALUE> ContinuousScale<VALUE, Double>.computeZoom(originScale: ContinuousScale<VALUE, Double>): DirectedZoom {
    if (range != originScale.range){
        logger.warn("Cannot compute the zoom for 2 scales with different ranges:: $this $originScale")
        return DirectedZoom()
    }

    val firstActual = this(domain.first())
    val firstOrigin = originScale(domain.first())

    val lastActual = this(domain.last())
    val lastOrigin = originScale(domain.last())

    val domainScaled = lastActual - firstActual
    val domainScaledOrigin = lastOrigin - firstOrigin

    val ratio = (domainScaled / domainScaledOrigin)
    val offset = this(originScale.domain.first()) - (range.first() * ratio)

    return DirectedZoom(ratio, offset)
}


