package io.data2viz.charts.dimension

import io.data2viz.charts.core.Datum
import kotlinx.datetime.Instant
import kotlinx.datetime.LocalDateTime

/**
 * [Temporal] is a [Dimension] that is used to manage continuous time values (as [LocalDateTime])
 * It is build by giving an "accessor" lambda that returns a [LocalDateTime] object from a given [Datum] of a <DOMAIN>.
 *
 * Use the [Chart.temporal] function to build a [Temporal] dimension.
 *
 * Examples:
 *  - val dateDimension = temporal( { domain.date } )
 */
public class Temporal<DOMAIN>(
    accessor: Datum<DOMAIN>.() -> Instant,
    parentDimension: Temporal<DOMAIN>? = null,
    init: Temporal<DOMAIN>.() -> Unit = {}
) : ContinuousDimension<DOMAIN, Instant>(
    accessor = accessor,
    parentDimension = parentDimension
) {

    init {
        init()
    }

    /**
     * Creates a child [Dimension] with the given accessor, use child dimensions to share axes with
     * different properties.
     *
     * Note: if current dimension is already a "child" dimension, attach this dimension to the current's parent.
     *
     * @param accessor This child [Dimension] accessor
     * @param init The initialization lambda for this [Dimension]
     */
    public fun child(accessor: Datum<DOMAIN>.() -> Instant, init: Temporal<DOMAIN>.() -> Unit = {}): Temporal<DOMAIN> {
        return Temporal(accessor, (parentDimension ?: this) as Temporal<DOMAIN>, init)
    }

}
