package io.data2viz.charts.event

import io.data2viz.charts.chart.mark.RelativePoint
import io.data2viz.math.Percent
import io.data2viz.math.pct

/**
 * A zoom action from chart-relative coordinates with a relative factor, defaults to 100% / 100% (no zoom).
 *
 * >100% = zoom-in
 * <100% = zoom-out.
 */
public data class ZoomAction internal constructor(
    public val zoomOrigin: RelativePoint,
    public val zoomX: Percent,
    public val zoomY: Percent,
    internal val isFromSelection: Boolean,
    internal val isFromPinch: Boolean,
    internal val pan: PanAction = PanAction()
) {
    public constructor(
        zoomOrigin: RelativePoint = RelativePoint(50.pct, 50.pct),
        zoomX: Percent = 100.pct,
        zoomY: Percent = 100.pct,
    ): this(zoomOrigin, zoomX, zoomY, false, false)

    /**
     * Get a copy of the current zoom with a X-axis constraint (this means that the zoom ratio in X is always 100%)
     */
    public fun constraintOnX(): ZoomAction = copy(zoomX =  100.pct)

    /**
     * Get a copy of the current zoom with a Y-axis constraint (this means that the zoom ratio in Y is always 100%)
     */
    public fun constraintOnY(): ZoomAction = copy(zoomY =  100.pct)
}

/**
 * ZoomActionFromSelection is a ZoomAction internally used to describe the fact that we zoom from a
 * selection rectangle, it should be handled differently when computing the zoom.
 *
 * If the zoom is only in X/Y then the origin of the ZoomAction will be computed differently.
 */
internal fun ZoomActionFromSelection(selectionCenter: RelativePoint, zoomX: Percent, zoomY: Percent) =
    ZoomAction(RelativePoint(50.pct - selectionCenter.x, 50.pct - selectionCenter.y), zoomX, zoomY, true, false)

/**
 * ZoomActionFromSelection is a ZoomAction internally used to describe the fact that we zoom from a
 * selection rectangle, it should be handled differently when computing the zoom.
 *
 * If the zoom is only in X/Y then the origin of the ZoomAction will be computed differently.
 */
internal fun ZoomActionFromPinch(selectionCenter: RelativePoint, zoomX: Percent, zoomY: Percent, pan: PanAction) =
    ZoomAction(selectionCenter, zoomX, zoomY, false, true, pan)
